#include <avr/io.h>

#define F_CPU 32000000

#include <asf.h>
/*------------------------------------------------*/
/* FFTEST : A test program for FFT module         */

#include <string.h>

#include <avr/pgmspace.h>
#include "ffft.h"		/* Defs for using Fixed-point FFT module */

#include "mydefs.h"
#include <util/delay.h>

#include <asf.h>
#include <stdlib.h>

/*------------------------------------------------*/
/* Global variables                               */

int16_t capture[FFT_N];
complex_t bfly_buff[FFT_N];
uint16_t spektrum[FFT_N/2];
uint16_t pwm_setting[6];
uint8_t counter[6];
volatile uint8_t peak[6];
char strng[20];
// xmega
int16_t ADC_result; //  fr ADC-Messung bentigt
void capture_wave(int16_t *buffer, uint16_t count);

uint16_t randNumber; // Nur fr Debug-Zwecke, zum Erzeugen von Random-Messwerte


const uint16_t pwmtable_10[64]  PROGMEM = { 0, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 4, 4, 5,
	5, 6, 6, 7, 8, 9, 10, 11, 12, 13, 15, 17,
	19, 21, 23, 26, 29, 32, 36, 40, 44, 49, 55,
	61, 68, 76, 85, 94, 105, 117, 131, 146, 162,
	181, 202, 225, 250, 279, 311, 346, 386, 430,
	479, 534, 595, 663, 739, 824, 918, 1023 };


// Messungen aus ADC aufnehmen
// in den *buffer mssen ADC-Werte zwischen -32768 und 32767 stehen. Dabei entsprechen -32768=0V und 32767=5V (im Original). 
// Hier wird es so hingebogen, dass es auf 3.3V zutrifft.
void capture_wave(int16_t *buffer, uint16_t count) {
	
	do {		
		ADCA.CH0.CTRL |= ADC_CH_INPUTMODE_DIFF_gc;
		ADCA.CH0.MUXCTRL |= ADC_CH_MUXPOS_PIN4_gc;
		ADCA.CH0.MUXCTRL |= ADC_CH_MUXNEG_PIN0_gc;
		ADCA.CH0.CTRL |= ADC_CH_START_bm;
		while(!ADCA.CH0.INTFLAGS);
		ADC_result= ADCA.CH0.RES; // ADC_result liegt zwischen 0 und 2^16 (bei 2.06V, da Referenz)

		// ADC_result wird so skaliert (1.3026) und verschoben (-34421.615), damit folgendes herauskommt:
		// 0V = -32768; 1.65V = 0; 3.3V = 32767
		// Hier kann man glaub ich noch was verbessern.
		ADC_result= (int16_t) Max(Min((((ADC_result)*1.3026155f)-34421.6155f)*2.0f,32766.0f),-32768.0f);

		// Nur fr Debug-Zwecke	
		//randNumber = (uint16_t) rand()*2;
		//ADC_result = (int16_t) (randNumber-32768);
	
			*buffer++ = ADC_result;	 //ADC_result - 32768
	} while(--count);
	
}

/*------------------------------------------------*/

ISR(TCC0_OVF_vect) {
	static uint16_t pwm_cnt = 0;
	uint8_t tmp = 0;

	// xmega
	//TCC0.PER += (uint16_t) T_PWM; // Original .. das fhrt aber dazu, dass die LED nur ab und zu blinken. Warum? Kein Ahnung
	TCC0.PER = (uint16_t) T_PWM;

	if(pwm_setting[0] > pwm_cnt) tmp |= Led1;
	if(pwm_setting[1] > pwm_cnt) tmp |= Led2;
	if(pwm_setting[2] > pwm_cnt) tmp |= Led3;
	if(pwm_setting[3] > pwm_cnt) tmp |= Led4;
	if(pwm_setting[4] > pwm_cnt) tmp |= Led5;
	if(pwm_setting[5] > pwm_cnt) tmp |= Led6;

	PORTA.OUT = tmp;

	if(pwm_cnt == (PWM_STEPS - 1))
	pwm_cnt = 0;
	else
	pwm_cnt++;

	// muss aus Performancegrnden leider mit in die ISR
	if((pwm_cnt == 512) || (pwm_cnt == 0)) {
		if(counter[0] == 0)
		if(peak[0]) peak[0]--;									// LED's zurckfaden

		if(counter[1] == 0)
		if(peak[1]) peak[1]--;

		if(counter[2] == 0)
		if(peak[2]) peak[2]--;

		if(counter[3] == 0)
		if(peak[3]) peak[3]--;

		if(counter[4] == 0)
		if(peak[4]) peak[4]--;

		if(counter[5] == 0)
		if(peak[5]) peak[5]--;

}
	
}


int main (void)
{
uint16_t s;	


// Insert system clock initialization code here (sysclk_init()).

// Ausgnge fr die 6 LEDs
PORTA.DIR |= (Led1 | Led2 | Led3 | Led4 | Led5 | Led6);	
	
	// initialize ADC
	PORTA.DIR &=~(1<<0); // kann sein, dass man das nicht braucht
	PORTA.DIR |=(1<<1); // wird glaub ich nicht bentigt. War fr Testzwecke
	PORTA.OUT |=(1<<1); // wird glaub ich nicht bentigt. War fr Testzwecke
	ADCA.CTRLB |= ADC_RESOLUTION_LEFT12BIT_gc | ADC_CONMODE_bm;	// 12 bit left adjust signed
	ADCA.REFCTRL |= ADC_REFSEL_INTVCC_gc;				// Vcc/1.6 = 2.06 V
	ADCA.PRESCALER |= ADC_PRESCALER_DIV128_gc;			// Prescaler von 128
	ADCA.CTRLA |= ADC_ENABLE_bm;						// ADC enablen

//Oszillator auf 32Mhz stellen
OSC.CTRL |= OSC_RC32MEN_bm;
// Warten bis der Oszillator bereit ist
while(!(OSC.STATUS & OSC_RC32MRDY_bm));
//Schtzt I/O Register, Interrupts werden ignoriert
CCP = CCP_IOREG_gc;
//aktiviert den internen Oszillator
CLK.CTRL = CLK_SCLKSEL_RC32M_gc;

// xmega
TCC0.CTRLA = TC_CLKSEL_DIV2_gc; // Systemtakt/2
TCC0.CTRLB = 0x00; // Normal-Modus
TCC0.INTCTRLA = 0x03; // Hchste Prioritt

sei();
// xmega
PMIC.CTRL |= PMIC_LOLVLEN_bm | PMIC_MEDLVLEN_bm | PMIC_HILVLEN_bm; // Interrupts freigeben

for(;;) {
	
	capture_wave(capture, FFT_N);
	fft_input(capture, bfly_buff);
	fft_execute(bfly_buff);
	fft_output(bfly_buff, spektrum);
	
	// Array in bentigter Gre vor Wiederverwendung 'subern'
	memset(capture, 0, 12);

	// Berechnung der einzelnen Bnder, ohne zunchst offensichtlichem System,
	// ergab sich so aus den Unzulnglichkeiten des AD-Wandlers,
	// (hab's versucht, mit rosa Rauschen so in etwa hinzubiegen)
	

	for (uint8_t n = 0; n < FFT_N / 2; n++) {
		s = spektrum[n];

		if(n == 0) {											// tiefstes Frequenzband
			if(s <= SCORE_MAX) capture[0] = s;
			else capture[0] = SCORE_MAX;
		}

		if((n >= 1)  && (n <= 3)) {
			if((capture[1] + s) <= SCORE_MAX) capture[1] += s;
			else capture[1] = SCORE_MAX;
		}

		if((n >= 4)  && (n <= 11)) {
			if((capture[2] + s) <= SCORE_MAX) capture[2] += s;
			else capture[2] = SCORE_MAX;
		}

		if((n >= 12) && (n <= 21)) {
			if((capture[3] + s) <= SCORE_MAX) capture[3] += s;
			else capture[3] = SCORE_MAX;
		}

		if((n >= 22) && (n <= 34)) {
			if((capture[4] + s) <= SCORE_MAX) capture[4] += s;
			else capture[4] = SCORE_MAX;
		}

		if((n >= 35) && (n <= 63)) {							// hchstes Frequenzband 35 63
			if((capture[5] + s) <= SCORE_MAX) capture[5] += s;
			else capture[5] = SCORE_MAX;
		}
	}

	for(uint8_t y = 0; y < 6; y++) {
		if(capture[y] > THRESHOLD) capture[y] -= THRESHOLD;		// Schwellwert abziehen
		else capture[y] = 0;

		capture[y] /= 256;										// Ergebnis teilen

		if(capture[y] > 63) capture[y] = 63;					// auf 64 Stufen begrenzen (Murphy's Law)

		if(counter[y]) counter[y]--;							// vor dem zurckfaden der LED's etwas warten

		if(capture[y] >= peak[y]) {								// so 'ne Art PeakHold
			peak[y] = capture[y];
			counter[y] = 3;
		}

		cli();
		pwm_setting[y] = pgm_read_word(&pwmtable_10[peak[y]]);	// Helligkeitsstufe zuweisen
		
		// DEBUG
		//pwm_setting[y] = 100;
		
		sei();
	}
	

}
}
